/*

*/

package tokenizer;

import java.util.List;
import java.util.ArrayList;

/**
   This {@code Tokenizer} splits a {@link String}
   into a {@link List} of {@link Token} objects.
   All the tokens from this tokenizer are either
   one or two non-space characters.
   <p>
   This tokenizer is split into two parts.
   <p>
   The first part is a private part that tokenizes
   a given input {@link String} into a {@link List}
   of {@link Token} objects. The private part of this
   tokenizer is only used by the {@code Tokenizer}
   constructor.
   <p>
   The second part of this tokenizer is a public part
   that lets a client program parse the {@link List}
   of {@link Token} objects that was built by the
   private part (the constructor).
   <p>
   This tokenizer is adapted from the online book
   "Crafting Interpreters", by Robert Nystrom.
   <p>
   See <a href="https://craftinginterpreters.com/scanning.html" target="_top">
                https://craftinginterpreters.com/scanning.html</a>
*/
public class Tokenizer
{
   // tokenizing state (used by the constructor)
   private final String source;
   private int currentChar;
   private int startLexeme;

   // tokenizer state (used by the public interface)
   private final List<Token> tokens = new ArrayList<>();
   private int currentToken;

   /**
      This constructor uses the private part of
      this {@code Tokenizer} to build the {@link List}
      of {@link Token} objects from the given input
      {@link String}.

      @param source  {@link String} to be tokenized
      @throws TokenizeError if there is an error in the string
   */
   public Tokenizer(final String source)
   {
      this.source = source;
      scanTokens();
      currentToken = 0;
   }

   /**
      Break up the input {@link String} into
      individual one or two non-space character
      units that are each represented by a
      {@link Token} object.
   */
   private void scanTokens()
   {
      currentChar = 0; // start at the beginning of the input string
      while ( ! isAtEnd() )
      {
         startLexeme = currentChar;
         scanToken(); // consume one token from the input string
      }
   }

   /**
      Consume one character from the input {@link String}
      and add its {@link Token} to the {@link List} of
      {@link Token} objects.
   */
   private void scanToken()
   {
      final char c = advance();
      switch (c)
      {
         // one character tokens
         case '(': addToken(); break;
         case ')': addToken(); break;
         case '+': addToken(); break;
         case '-': addToken(); break;
         case '/': addToken(); break;
         case '%': addToken(); break;
         case '0': addToken(); break;
         case '1': addToken(); break;
         case '2': addToken(); break;
         case '3': addToken(); break;
         case '4': addToken(); break;
         case '5': addToken(); break;
         case '6': addToken(); break;
         case '7': addToken(); break;
         case '8': addToken(); break;
         case '9': addToken(); break;

         // one or two character tokens
         case '*': match('*'); addToken(); break;
         case '|': match('|'); addToken(); break;
         case '&': match('&'); addToken(); break;
         case '=': match('='); addToken(); break;
         case '!': match('='); addToken(); break;
         case '<': match('='); addToken(); break;
         case '>': match('='); addToken(); break;

         // whitespace
         case ' ':
         case '\r':
         case '\t':
         case '\n':
           // Ignore whitespace.
           break;

         default: // error
           throw new TokenizeError("unexpected character: "
                                 + "'" + c + "'"
                                 + "at position " + startLexeme);
      }
   }

   private void addToken()
   {
      final String text = source.substring(startLexeme, currentChar);
      tokens.add( new Token(text, startLexeme) );
   }

   private char advance()
   {
      ++currentChar;
      return source.charAt(currentChar - 1);
   }

   private boolean match(final char expected)
   {
      if (isAtEnd()) return false;
      if (source.charAt(currentChar) != expected) return false;
      ++currentChar;
      return true;
   }

   private boolean isAtEnd()
   {
      return currentChar >= source.length();
   }


   // The following methods are the public interface to the tokenizer.

   /**
      Return true if all of the tokens have been consumed.

      @return true if all of the tokens have been consumed
   */
   public boolean isEmpty()
   {
      return currentToken >= tokens.size();
   }

   /**
      Return true if there are tokens that have not yet been consumed.

      @return true if there are tokens that have not been consumed
   */
   public boolean hasToken()
   {
      return currentToken < tokens.size();
   }


   /**
      Return the number of unconsumed tokens.

      @return the number of unconsumed tokens
   */
   public int size()
   {
      return tokens.size() - currentToken;
   }


   /**
      This method allows you to look at the current
      token without consuming it.

      @return the current {@link Token} without consuming it
      @throws TokenizeError if there are no remaining tokens
   */
   public Token peek()
   {
      if ( isEmpty() ) throw new TokenizeError("no more tokens");

      return tokens.get(currentToken);
   }


   /**
      This method consumes and returns the current token.

      @return the next {@link Token}
      @throws TokenizeError if there are no remaining tokens
   */
   public Token next()
   {
      if ( isEmpty() ) throw new TokenizeError("no more tokens");

      return tokens.get(currentToken++);
   }


   /**
      This method both tests and consumes the current token.

      @param lexeme {@link String} to be compared with the next {@link Token}
      @return true if {@code lexeme} matches the next {@link Token}
      @throws TokenizeError if there are no remaining tokens
   */
   public boolean match(final String lexeme)
   {
      if ( isEmpty() ) throw new TokenizeError("no more tokens");

      final boolean match =
             lexeme.equals( tokens.get(currentToken).lexeme );

      if (match)
      {
         ++currentToken; // consume the matched token
         return true;
      }
      else
      {
         return false; // don't consume the unmatched token
      }
   }


   @Override
   public String toString()
   {
      String result = "";
      for (final Token t : tokens)
      {
         result += t + "\n";
      }
      return result;
   }
}
